/* auditctl.c -- 
 * Copyright 2004, 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *     Steve Grubb <sgrubb@redhat.com>
 *     Rickard E. (Rik) Faith <faith@redhat.com>
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#define __USE_XOPEN_EXTENDED
#define __USE_GNU
#include <string.h>	/* strdup needs xopen define */
#undef __USE_GNU
#undef __USE_XOPEN_EXTENDED
#include <getopt.h>
#include <time.h>
#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/utsname.h>
#define __USE_GNU
#include <fcntl.h>	/* O_NOFOLLOW needs gnu defined */
#undef __USE_GNU
#include <errno.h>
#include <libgen.h>	/* For basename */
#include "libaudit.h"


/* This define controls how many rule options we will allow when
 * reading a rule from a file. 64 fields are allowed by the kernel, so I
 * want to allow that plus a few entries for lists and other such items */
#define NUM_OPTIONS 72

/* This define controls the size of the line that we will request when
 * reading in rules from a file. We need to allow 64 fields. 25 bytes is 
 * the largest syscall name, so lets allow 1600 per line. 
 * Unrealistic - I know. 
 */
#define LINE_SIZE 1600

/* These are all related to the fs watch code */
#define WATCH_MAY_EXEC		1
#define WATCH_MAY_WRITE		2
#define WATCH_MAY_READ		4
#define WATCH_MAY_APPEND	8


/* Global functions */
static int handle_request(int status);
static void get_reply(void);
static int audit_print_reply(struct audit_reply *rep);
static int delete_all_rules(void);

/* Global vars */
static int fd = -1;
static int list_requested = 0;
static int add = 0, del = 0, action = 0;
static int ins = 0, rem = 0;
static struct audit_rule  rule;
static struct audit_watch watch;

extern int audit_archadded;
extern int audit_syscalladded;
extern unsigned int audit_elf;

/*
 * This function will reset everything used for each loop when loading 
 * a ruleset from a file.
 */
static int reset_vars(void)
{
	list_requested = 0;
	audit_syscalladded = 0;
	audit_archadded = 0;
	audit_elf = 0;
	add = 0;
	del = 0;
	action = 0;
	ins = 0;
	rem = 0;

	memset(&rule, 0, sizeof(rule));
	memset(&watch, 0, sizeof(watch));
	if ((fd = audit_open()) < 0) {
		fprintf(stderr, "Cannot open netlink audit socket\n");
		return 1;
	}
	return 0;
}

static void usage(void)
{
    printf("usage: auditctl [options]\n"
     "       -a <l,a>     Append rule to end of <l>ist with <a>ction\n"
     "       -A <l,a>     Add rule at beginning of <l>ist with <a>ction\n"
     "       -b <backlog> Set max number of outstanding audit buffers allowed\n"
     "                    Default=64\n"
     "       -d <l,a>     Delete rule from <l>ist with <a>ction\n"
     "                    l=task,entry,exit a=never,possible,always\n"
     "       -D           Delete all rules\n"
     "       -e [0|1]     Set enabled flag\n"
     "       -f [0..2]    Set failure flag\n"
     "                    0=silent 1=printk 2=panic\n"
     "       -F f=v       Build rule: field name, value\n"
     "       -h           Help\n"
     "       -k <key>     Set filterkey on watch\n"
     "       -l           List rules\n"
     "       -m text      Send a user-space message\n"
     "       -p [r|w|e|a] Set permissions filter on watch\n"
     "                    r=read, w=write, e=execute, a=append\n"
     "       -r <rate>    Set limit in messages/sec (0=none)\n"
     "       -R <file>    read rules from file\n"
     "       -s           Report status\n"
     "       -S syscall   Build rule: syscall name or number\n"
     "       -t <syscall> Translate syscall number to syscall name\n"
     "       -v           Version\n"
     "       -w <path>    Insert watch at <path>\n"
     "       -W <path>    Remove watch at <path>\n"
     );
}

static int audit_rule_setup(const char *opt, int *flags, int *act)
{
	if (strstr(opt, "task")) 
		*flags = AUDIT_PER_TASK;
	else if (strstr(opt, "entry"))
		*flags = AUDIT_AT_ENTRY;
	else if (strstr(opt, "exit"))
		*flags = AUDIT_AT_EXIT;
	else
		return 1;
	if (strstr(opt, "never"))
		*act = AUDIT_NEVER;
	else if (strstr(opt, "possible"))
		*act = AUDIT_POSSIBLE;
	else if (strstr(opt, "always"))
		*act = AUDIT_ALWAYS;
	else
		return 1;
	return 0;
}

/*
 * This function will check the path before accepting it. It returns
 * 1 on error and 0 on success.
 */
static int check_path(const char *path)
{
	char *ptr, *base;
	int len = strlen(path);
	if (len > PATH_MAX) {
		fprintf(stderr, "The path passed for the watch is too big\n");
		return 1;
	}
	ptr = strdup(path);
	base = basename(ptr);
	len = strlen(base);
	free(ptr);
	if (len > NAME_MAX) {
		fprintf(stderr, "The base name of the path is too big\n");
		return 1;
	}

	/* These are warnings, not errors */
	if (strstr(path, ".."))
		fprintf(stderr, 
			"Warning - relative path notation is not supported");
	if (strchr(path, '*'))
		fprintf(stderr, "Warning - wildcard notation is not supported");
	if (strchr(path, '?'))
		fprintf(stderr, "Warning - wildcard notation is not supported");

	return 0;
}

/*
 * Setup a watch.  The "name" of the watch in userspace will be the <path> to
 * the watch.  When this potential watch reaches the kernel, it will resolve
 * down to <name> (of terminating file or directory). 
 * Returns a 1 on success & -1 on failure.
 */
static int audit_setup_watch_name(struct audit_watch *req, const char *opt, 
	int *act)
{
	if (!req->name) {
		if (check_path(opt))
			return -1;
		req->name = strdup(opt);
		if (!req->name) {
			fprintf(stderr, "Out of memory\n");
			return -1;
		}
		req->namelen = strlen(req->name) + 1;
		*act = 1;
		return 1;
	}
	fprintf(stderr, "There's already a watch given\n");
	return -1;
}

/*
 * Setup a filterkey for the watch.  
 * Returns a 1 on success & -1 on failure.
 */
static int audit_setup_filterkey(struct audit_watch *req, const char *opt)
{
	if (!req->filterkey) {
		req->filterkey = strdup(opt);
		if (!req->filterkey) {
			fprintf(stderr, "Out of memory\n");
			return -1;
		}
		req->fklen = strlen(opt) + 1;
		if (req->fklen > AUDIT_FILTERKEY_MAX) {
			fprintf(stderr, "The filterkey is too big\n");
			free(req->filterkey);
			return -1;
		}
		return 1;
	}
	fprintf(stderr, "There's already a filterkey given\n");
	return -1;
}

/*
 * Setup a watch permissions.
 * Returns a 1 on success & -1 on failure.
 */
static int audit_setup_perms(struct audit_watch *req, const char *opt)
{
	unsigned int i;

	if (strlen(opt) > 4)
		return -1;

	for (i = 0; i < strlen(opt); i++) {
		switch (tolower(opt[i])) {
			case 'r':
				if (!(req->perms & WATCH_MAY_READ))
					req->perms |= WATCH_MAY_READ;
				else
					goto audit_watch_setup_exit;
				break;
			case 'w':
				if (!(req->perms & WATCH_MAY_WRITE))
					req->perms |= WATCH_MAY_WRITE;
				else
					goto audit_watch_setup_exit;
				break;
			case 'e':
				if (!(req->perms & WATCH_MAY_EXEC))
					req->perms |= WATCH_MAY_EXEC;
				else
					goto audit_watch_setup_exit;
				break;
			case 'a':
				if (!(req->perms & WATCH_MAY_APPEND))
					req->perms |= WATCH_MAY_APPEND;
				else
					goto audit_watch_setup_exit;
				break;
			default:
				fprintf(stderr, 
					"Permission %c isn't supported\n", 
					opt[i]);
				return -1;
		}
	}
	return 1;

audit_watch_setup_exit:
	fprintf(stderr, "Permission %c was already given\n", opt[i]);
	return -1;
}

void audit_request_both_lists(int fd)
{
	if (audit_request_rules_list(fd)) {
		list_requested = 1;
		get_reply();
	} else 
		fprintf(stderr, "Error requesting rules list\n");
	if (audit_request_watch_list(fd)) {
		list_requested = 2;
		get_reply();
	} else
		fprintf(stderr, "Error requesting watches list\n");
}

/*
 * returns: < 0 error - noreply, 0 success - reply, > 0 success - rule
 */
static int setopt(int count, char *vars[])
{
    int c;
    int retval = 0;

    optind = 0;
    opterr = 0;
    while ((retval != -1) && (c = getopt(count, vars,
			"hslDve:f:r:b:a:A:d:S:F:m:t:R:w:W:k:p:")) != EOF) {
        switch (c) {
        case 'h':
		usage();
		retval = -1;
		break;
        case 's':
		retval = audit_request_status(fd);
		if (retval <= 0)
			retval = -1;
		else
			retval = 0;
		break;
        case 'e':
		if (optarg && ((optarg[0] == '0')||(optarg[0] == '1'))) {
			if (audit_set_enabled(fd, strtol(optarg,NULL,0)))
				audit_request_status(fd);
		} else {
			fprintf(stderr, "Enable must be 0 or 1 was %s\n", 
				optarg);
			retval = -1;
		}
		break;
        case 'f':
		if (optarg && ((optarg[0] == '0')||(optarg[0] == '1')||
				(optarg[0] == '2'))) {
			if (audit_set_failure(fd, strtol(optarg,NULL,0)))
				audit_request_status(fd);
		} else {
			fprintf(stderr, "Failure must be 0, 1, or 2 was %s\n", 
				optarg);
			retval = -1;
		}
		break;
        case 'r':
		if (optarg && isdigit(optarg[0])) { 
			if (audit_set_rate_limit(fd, strtol(optarg,NULL,0)))
				audit_request_status(fd);
		} else {
			fprintf(stderr, "Rate must be a numeric value was %s\n",
				optarg);
			retval = -1;
		}
		break;
        case 'b':
		if (optarg && isdigit(optarg[0])) {
			if (audit_set_backlog_limit(fd, strtol(optarg,NULL,0)))
				audit_request_status(fd);
		} else {
			fprintf(stderr, 
				"Backlog must be a numeric value was %s\n", 
				optarg);
			retval = -1;
		}
		break;
        case 'l':
		audit_request_both_lists(fd);
		retval = -1;
		break;
        case 'a':
		if (strstr(optarg, "task") && audit_syscalladded) {
			fprintf(stderr, 
				"Syscall auditing requested for task list\n");
			retval = -1;
		}
		else if (audit_rule_setup(optarg, &add, &action) != 0) {
			fprintf(stderr, "Append rule - bad keyword %s\n", 
				optarg);
			retval = -1;
		}
		else
			retval = 1; /* success - please send */
		break;
        case 'A': 
		if (strstr(optarg, "task") && audit_syscalladded) {
			fprintf(stderr, 
				"Syscall auditing requested for task list\n");
			retval = -1;
		}
		else if (audit_rule_setup(optarg, &add, &action) == 0) {
			add |= AUDIT_PREPEND;
			retval = 1; /* success - please send */
		}
		else {
			fprintf(stderr, "Add rule - bad keyword %s\n", optarg);
			retval = -1;
		}
		break;
        case 'd': 
		if (audit_rule_setup(optarg, &del, &action) != 0) {
			fprintf(stderr, "Delete rule - bad keyword %s\n", 
				optarg);
			retval = -1;
		}
		else
			retval = 1; /* success - please send */
		break;
        case 'S':
		if ((add & AUDIT_PER_TASK || del & AUDIT_PER_TASK)) {
			fprintf(stderr, 
				"Syscall auditing being added to task list\n");
			return -1;
		} else {
			if (!audit_elf) {
				int machine;
				unsigned int elf;
				machine = audit_detect_machine();
				if (machine < 0) {
					fprintf(stderr, 
					    "Error detecting machine type");
					return -1;
				}
				elf = audit_machine_to_elf(machine);
                                if (elf == 0) {
					fprintf(stderr, 
					    "Error looking up elf type");
					return -1;
				}
				audit_elf = elf;
			}
		}
		switch (audit_rule_syscallbyname(&rule, optarg))
		{
			case 0:
				audit_syscalladded = 1;
				break;
			case -1:
				fprintf(stderr, "Syscall name unknown: %s\n", 
							optarg);
				retval = -1;
				break;
			case -2:
				fprintf(stderr, "Elf type unknown: 0x%x\n", 
							audit_elf);
				retval = -1;
				break;
		}
		break;
        case 'F': 
		switch (audit_rule_fieldpair(&rule, optarg))
		{
			case 0:
				break;
			case -1:
				fprintf(stderr, "-F missing = for %s\n", 
					optarg);
				retval = -1;
				break;
			case -2:
				fprintf(stderr, "-F unknown field: %s\n", 
					optarg);
				retval = -1;
				break;
			case -3:
				fprintf(stderr, 
					"-F %s must be before -S\n", 
					optarg);
				retval = -1;
				break;
			case -4:
				fprintf(stderr, 
					"-F %s machine type not found\n", 
					optarg);
				retval = -1;
				break;
			case -5:
				fprintf(stderr, 
					"-F %s elf mapping not found\n", 
					optarg);
				retval = -1;
				break;
			case -6:
				fprintf(stderr, 
					"-F %s requested bit level not supported by machine\n", 
					optarg);
				retval = -1;
				break;
			default:
				retval = -1;
				break;
		}
		break;
        case 'm':
		if (audit_send_message(fd, AUDIT_USER, optarg) <= 0)
			retval = -1;
		break;
        case 't':
		if (optarg && isdigit(optarg[0])) {
			const char *ptr, *mstr;
			int machine;
			if (audit_elf) {
				machine = audit_elf_to_machine(audit_elf);
				if (machine < 0) {
					fprintf(stderr, 
					    "Error translating machine type");
					return -1;
				}
			} else {
				machine = audit_detect_machine();
				if (machine < 0) {
					fprintf(stderr, 
					    "Error detecting machine type");
					return -1;
				}
			}
			mstr = audit_machine_to_name(machine);
			if (mstr == NULL)
				mstr = "Unknown";
                        ptr = audit_syscall_to_name(strtol(optarg, NULL,0), machine);
			if (ptr)
				printf("arch:%s, syscall:%s\n", mstr, ptr);
			else
				printf("Translate unknown value: %s\n", optarg);
		}
		else  
			fprintf(stderr, 
				"Translate needs a numeric value was %s\n",
				 optarg);
		retval = -1; /* No reply needed */
		break;
	case 'R':
		fprintf(stderr, "Error - nested rule files not supported\n");
		retval = -1;
		break;
	case 'D':
		retval = delete_all_rules();
		break;
	case 'w':
		if (optarg) 
			retval = audit_setup_watch_name(&watch, optarg, &ins);
		else {
			fprintf(stderr, "watch option needs a path\n");	
			retval = -1;
		}
		break;
	case 'W':
		if (optarg) 
			retval = audit_setup_watch_name(&watch, optarg, &rem);
		else {
			fprintf(stderr, "watch option needs a path\n");	
			retval = -1;
		}
		break;
	case 'k':
		if (!ins && !rem) {
			fprintf(stderr, 
			"filterkey option needs a watch given prior to it\n");	
			retval = -1;
		}
		else if (!optarg) {
			fprintf(stderr, "filterkey option needs a filter\n");	
			retval = -1;
		}
		else 
			retval =  audit_setup_filterkey(&watch, optarg);
		break;
	case 'p':
		if (!ins && !rem) {
			fprintf(stderr, 
			"permission option needs a watch given prior to it\n");	
			retval = -1;
		}
		else if (!optarg) {
			fprintf(stderr, "permission option needs a filter\n");	
			retval = -1;
		}
		else
			retval = audit_setup_perms(&watch, optarg);
		break;
	case 'v':
			printf("auditctl version %s\n", VERSION);
			break;
        default: 
		usage();
		retval = -1;
		break;
        }
    }
    /* This is to ignore errors where the user types "- s" */
    if (optind == 1)
	retval = -1;
    return retval;
}

static char *get_line(FILE *f, char *buf)
{
	if (fgets(buf, LINE_SIZE, f)) {
		/* remove newline */
		char *ptr = strchr(buf, 0x0a);
		if (ptr)
			*ptr = 0;
		return buf;
	}
	return NULL;
}

/*
 * This function reads the given file line by line and executes the rule.
 * It returns 0 if everything went OK, 1 if there are problems before reading
 * the file and -1 on error conditions after executing some of the rules.
 * It will abort reading the file if it encounters any problems.
 */
static int fileopt(const char *file)
{
	int i, tfd, rc, lineno = 1;
	struct stat st;
        FILE *f;
        char buf[LINE_SIZE];

	/* Does the file exist? */
	rc = open(file, O_NOFOLLOW|O_RDONLY);
	if (rc < 0) {
		if (errno != ENOENT) {
			fprintf(stderr,"Error opening %s (%s)\n", 
				file, strerror(errno));
                        return 1;
                }
                fprintf(stderr, "file %s doesn't exist, skipping\n", file);
                return 0;
        }
        tfd = rc;

	/* Is the file permissions sane? */
	if (fstat(tfd, &st) < 0) {
		fprintf(stderr, "Error fstat'ing %s (%s)\n",
			file, strerror(errno));
		close(tfd);
		return 1;
	}
	if (st.st_uid != 0) {
		fprintf(stderr, "Error - %s isn't owned by root\n", file);
		close(tfd);
		return 1;
	} 
	if ((st.st_mode & S_IWOTH) == S_IWOTH) {
		fprintf(stderr, "Error - %s is world writable\n", file);
		close(tfd);
		return 1;
	}
	if (!S_ISREG(st.st_mode)) {
		fprintf(stderr, "Error - %s is not a regular file\n", file);
		close(tfd);
		return 1;
	}

        f = fdopen(tfd, "r");
        if (f == NULL) {
                fprintf(stderr, "Error - fdopen failed (%s)\n",
                        strerror(errno));
		close(tfd);
                return 1;
        }

	/* Read until eof, lineno starts as 1 */
	while (get_line(f, buf)) {
		char *options[NUM_OPTIONS];
		char *ptr;

		/* Weed out blank lines */
		if (buf[0] == 0) {
			lineno++;
			continue;
		}
		ptr = strtok(buf, " ");
		if (ptr == NULL)
			break;
		/* allow comments */
		if (ptr[0] == '#') {
			lineno++;
			continue;
		}
		i = 0;
		options[i++] = "auditctl";
		options[i++] = ptr;
		while( (ptr=strtok(NULL, " ")) && i<NUM_OPTIONS-1 ) {
			options[i++] = ptr;
		}
		options[i] = NULL;

		/* Parse it */
		if (reset_vars()) {
			fclose(f);
			return -1;
		}
		rc = setopt(i, options);

		/* handle reply or send rule */
		if (handle_request(rc) == -1) {
			fprintf(stderr, "There was an error in line %d of %s\n",
				lineno, file);
			fclose(f);
			return -1;
		}
		lineno++;
	}
	fclose(f);
	return 0;
}

int main(int argc, char *argv[])
{
	int retval = 1;

	/* Check where the rules are coming from: commandline or file */
	if ((argc == 3) && (strcmp(argv[1], "-R") == 0)) {
#ifndef DEBUG
		/* Make sure we are root */
		if (getuid() != 0) {
			fprintf(stderr, 
				"You must be root to run this program.\n");
			return 4;
		}
#endif
		if (fileopt(argv[2]))
			return 1;
		else
			return 0;
	} else {
		if (argc == 1) {
			usage();
			return 1;
		}
#ifndef DEBUG
		/* Make sure we are root */
		if (getuid() != 0) {
			fprintf(stderr, 
				"You must be root to run this program.\n");
			return 4;
		}
#endif
		if (reset_vars())
			return 1;
		retval = setopt(argc, argv);
	}

	return handle_request(retval);
}

/*
 * This function is called after setopt to handle the return code.
 * status = 0 means just get the reply. Greater than 0 means we
 * are adding or deleting a rule or watch. Less than 0 means an error
 * occurred. Even if there's an error, we need to call this routine 
 * to close up the audit fd.
 */
static int handle_request(int status)
{
	if (status == 0)
		get_reply();
	else if (status > 0) {
		int rc;
		if (add & 0x07) 
			rc = audit_add_rule(fd, &rule, add, action);
		else if (del & 0x07) 
			rc = audit_delete_rule(fd, &rule, del, action);
		else if (ins && !rem)
			rc = audit_insert_watch(fd, &watch);
		else if (rem && !ins)
			rc = audit_remove_watch(fd, &watch);
		else {
        		usage();
	    		audit_close(fd);
			exit(1);
	    	}
		if (rc > 0) {
			if (audit_request_rules_list(fd) > 0) {
				list_requested = 1;
				get_reply();
				status = 0;
				if (audit_request_watch_list(fd) > 0) {
					list_requested = 2;
					get_reply();
				}
			} else {
				fputs("Error requesting rules list\n", stderr);
				status = -1;
			}
		} else {
			if (add || del)
				fputs("Error sending rule to kernel\n",
					stderr);
			else
				fputs("Error sending watch to kernel\n",
					stderr);
			status = -1;
		}
	} else 
		status = -1;

	audit_close(fd);
	fd = -1;
	return status;
}

/*
 * A reply from the kernel is expected. Get and display it.
 */
static void get_reply(void)
{
	int i, retval;
	int timeout = 30; /* loop has delay of .1 - so this is 3 seconds */
	struct audit_reply rep;

	for (i = 0; i < timeout; i++) {
		retval = audit_get_reply(fd, &rep, GET_REPLY_NONBLOCKING, 0);
		if (retval > 0) {
			struct timespec t;

			if (rep.type == NLMSG_ERROR && rep.error->error == 0)
				continue;
			t.tv_sec  = 0;
			t.tv_nsec = 100000000; /* .1 second */
			nanosleep(&t, NULL);
			if ((retval = audit_print_reply(&rep)) == 0) { 
				retval = 0;
				break;
			}
			else if (list_requested && retval == 1)
				i = 0; /* If getting a list, reset timeout */
		}
	}
}

struct audit_watch *transport_to_watch(void *memblk)
{
	unsigned int offset;
	struct watch_transport *r = (struct watch_transport *)memblk;
	struct audit_watch *req = (struct audit_watch *)malloc(
					sizeof(struct audit_watch));
	if (req == NULL)
		return NULL;

	req->dev_major = r->dev_major;
	req->dev_minor = r->dev_minor;
	req->perms = r->perms;
	req->valid = r->valid;
	req->namelen = r->pathlen;
	req->fklen =  r->fklen;

	offset = sizeof(struct watch_transport);
	req->filterkey = strndup(memblk + offset, req->fklen);
	offset += req->fklen;
	req->name = strndup(memblk + offset, req->namelen);

	return req;
}

static int audit_print_reply(struct audit_reply *rep)
{
	unsigned int i;
	int first;
	int sparse;
	int machine = audit_detect_machine();

	audit_elf = 0; 
	switch (rep->type) {
		case NLMSG_NOOP:
			return 1;
		case NLMSG_DONE:
			if (list_requested == 1)
				printf("No rules\n");
			else if (list_requested == 2)
				printf("No watches\n");
			return 0;
		case NLMSG_ERROR: 
		        printf("NLMSG_ERROR %d (%s) type=%d seq=%d\n",
				-rep->error->error, 
				strerror(-rep->error->error),
				rep->type, rep->nlh->nlmsg_seq);
			return 0;
		case AUDIT_GET:
			printf("AUDIT_STATUS: enabled=%d flag=%d pid=%d"
			" rate_limit=%d backlog_limit=%d lost=%d backlog=%d\n",
			rep->status->enabled, rep->status->failure,
			rep->status->pid, rep->status->rate_limit,
			rep->status->backlog_limit, rep->status->lost,
			rep->status->backlog);
			return 0;
		case AUDIT_LIST:
			list_requested = 0;
			printf("AUDIT_LIST: %s %s",
			audit_flag_to_name((int)rep->rule->flags),
			audit_action_to_name(rep->rule->action));
			for (i = 0; i < rep->rule->field_count; i++) {
				int field = rep->rule->fields[i];
				int negated = 0;

				if (field & AUDIT_NEGATE) {
					field &= ~AUDIT_NEGATE;
					negated  = 1;
				}
                
				const char *name = audit_field_to_name(field);
				if (name) {
					if (strcmp(name, "arch") == 0) { 
						audit_elf =rep->rule->values[i];
						printf(" %s%s%u", name, 
							negated ? "!=" : "=",
							(unsigned)rep->rule->values[i]);
					}
					else 
						printf(" %s%s%d", name, 
							negated ? "!=" : "=",
							rep->rule->values[i]);
				}
				else 
					printf(" f%d%s%d", rep->rule->fields[i],
						negated ? "!=" : "=",
						rep->rule->values[i]);
				if (rep->rule->values[i])
					printf(" (0x%x)", rep->rule->values[i]);
			}
			printf(" syscall=");
			for (sparse = 0, i = 0; i < AUDIT_BITMASK_SIZE; i++)
				if (rep->rule->mask[i] != (unsigned int)~0)
					sparse = 1;
			if (!sparse) {
				printf("all");
			} else for (first = 1, i = 0;
					i < AUDIT_BITMASK_SIZE * 32; i++) {
				int word = AUDIT_WORD(i);
				int bit  = AUDIT_BIT(i);
				if (rep->rule->mask[word] & bit) {
					const char *ptr;
					if (audit_elf)
						machine = audit_elf_to_machine(
							audit_elf);
					if (machine < 0)
						ptr = 0;
					else
						ptr = audit_syscall_to_name(i, 
						machine);
					if (ptr)
						printf("%s%s", 
						first ? "" : ",", ptr);
					else
						printf("%s%d", 
						first ? "" : ",", i);
					first = 0;
				}
			}
			printf("\n");
			return 1; /* get more messages until NLMSG_DONE */
		case AUDIT_WATCH_LIST:
		{
			struct audit_watch *req =transport_to_watch(rep->watch);
			list_requested = 0;
			if (!req) {
				fputs("No memory receiving watch list\n", 
						stderr);
				return 0;
			}
			printf("AUDIT_WATCH_LIST: dev=%u:%u, path=%s, "
				"filterkey=%s, perms=%u, valid=%d\n",
				req->dev_major, req->dev_minor, req->name,
				req->filterkey, req->perms, req->valid);
			free(req->name);
			free(req->filterkey);
			free(req);
			return 1; /* get more messages until NLMSG_DONE */
		}
		default:
			printf("Unknown: type=%d, len=%d\n", rep->type, 
				rep->nlh->nlmsg_len);
			return 0;
	}
}

static int delete_all_rules(void)
{
	int rc, i;
	int timeout = 30; /* tenths of seconds */
	struct audit_reply rep;

	/* list the rules */
	rc = audit_request_rules_list(fd);
	if (rc <= 0)
		return -1;

	for (i = 0; i < timeout; i++) {
		struct timespec t;

		t.tv_sec  = 0;
		t.tv_nsec = 100000000; /* .1 second */
		nanosleep(&t, NULL);
		rc = audit_get_reply(fd, &rep, GET_REPLY_NONBLOCKING, 0);
		if (rc > 0) {
			/* If we get done or error, break out */
			if (rep.type == NLMSG_DONE )
				break;

			if (rep.type == NLMSG_ERROR) {
				printf("Error receiving list (%s)\n", 
					strerror(errno));
				return -1;
			}

			/* If its not what we are expecting, keep looping */
			if (rep.type != AUDIT_LIST)
				continue;

			/* Found it, bounce it right back with delete */
			rc = audit_send(fd, AUDIT_DEL, rep.rule, 
					sizeof(struct audit_rule));
			if (rc < 0) {
				printf("Error deleting rule\n"); 
				return -1;
			}

			/* Reset timeout */
			i = 0;
		}
	}
	
	/* Finish by proving we are empty */
	list_requested = 1;
	if (audit_request_rules_list(fd) < 0)
		return -1;
	else
		return 0;
}

