/* auditd.c -- 
 * Copyright 2004, 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *   Steve Grubb <sgrubb@redhat.com>
 *   Rickard E. (Rik) Faith <faith@redhat.com>
 */

#include "config.h"
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <sys/resource.h>
#include <sys/time.h>
#define __USE_GNU
#include <fcntl.h>
#include <pthread.h>
#undef __USE_GNU

#include "libaudit.h"
#include "auditd-config.h"
#include "auditd-event.h"

#define DMSG_SIZE 100

/* Global Data */
volatile int stop = 0;

/* Local data */
static int fd = -1;
static struct daemon_conf config;
static const char *pidfile = "/var/run/auditd.pid";

/* Local function prototypes */
static void clean_exit(void);
static int get_reply(int fd, struct audit_reply *rep);


/*
 * Output a usage message
 */
static void usage(void)
{
	puts("Usage: auditd [ -f ]");
	exit(2);
}


/*
 * SIGTERM handler
 */ 
static void term_handler( int sig )
{
	stop = 1;
}

static void thread_killer( int sig )
{
	exit(0);
}

/*
 * This function is used to send start, stop, and abort messages 
 * to the audit log. NOTE: This function leaks memory - use sparingly.
 */
static unsigned seq_num = 0;
static int send_audit_event(int type, const char *str)
{
	struct auditd_reply_list *rep;
	struct timeval tv;
	
	if ((rep = malloc(sizeof(*rep))) == NULL) {
		audit_msg(LOG_ERR, "Cannot allocate audit reply");
		return 1;
	}

	rep->reply.type = type;
	rep->reply.message = (char *)malloc(DMSG_SIZE); /* Here's the leak */
	if (rep->reply.message == NULL) {
		audit_msg(LOG_ERR, "Cannot allocate local event message");
		return 1;
	}
	if (seq_num == 0) {
		srand(time(NULL));
		seq_num = rand()%1000;
	} else
		seq_num++;
	if (gettimeofday(&tv, NULL) == 0) {
		rep->reply.len = snprintf((char *)rep->reply.message,
			DMSG_SIZE, "audit(%lu.%03u:%u) %s, auditd pid=%u", 
			tv.tv_sec, (unsigned)(tv.tv_usec/1000), seq_num, str,
			getpid());
	} else {
		rep->reply.len = snprintf((char *)rep->reply.message,
			DMSG_SIZE, "audit(%lu.%03u:%u) %s, auditd pid=%u", 
			(unsigned long)time(NULL), 0, seq_num, str, getpid());
	}
	if (rep->reply.len > DMSG_SIZE)
		rep->reply.len = DMSG_SIZE;

	enqueue_event(rep);
	pthread_yield(); /* Give the other thread a chance to log it. */

	return 0;
}

static int write_pid(void)
{
	int pidfd, len;
	char val[16];

	len = snprintf(val, sizeof(val), "%u\n", getpid());
	if (len < 0) {
		audit_msg(LOG_ERR, "Pid error (%s)", strerror(errno));
		pidfile = 0;
		return 1;
	}
	pidfd = open(pidfile, O_CREAT | O_TRUNC | O_NOFOLLOW | O_WRONLY, 0644);
	if (pidfd < 0) {
		audit_msg(LOG_ERR, "Unable to set pidfile (%s)", strerror(errno));
		pidfile = 0;
		return 1;
	}
	(void)write(pidfd, val, (unsigned int)len);
	close(pidfd);
	return 0;
}

int main(int argc, char *argv[])
{
	struct sigaction sa;
	fd_set read_mask;
	struct auditd_reply_list *rep = NULL;
	struct rlimit limit;

	/* Get params && set mode */
	if (argc > 2)
		usage();
	if (argc == 2) {
		if (strcmp(argv[1], "-f") == 0) {
			config.daemonize = D_FOREGROUND;
			set_aumessage_mode(MSG_STDERR, DBG_YES);
		} else
			usage();
	}
	else {
		config.daemonize = D_BACKGROUND;
		set_aumessage_mode(MSG_SYSLOG, DBG_NO);
	}

#ifndef DEBUG
	/* Make sure we are root */
	if (getuid() != 0) {
		fprintf(stderr, "You must be root to run this program.\n");
		return 4;
	}
#endif

	/* Register sighandlers */
	sa.sa_flags = 0 ;
	sa.sa_handler = term_handler;
	sigemptyset( &sa.sa_mask ) ;
	sigaction( SIGTERM, &sa, NULL );
	sa.sa_handler = SIG_IGN;
	sigaction( SIGHUP, &sa, NULL );
	sigaction( SIGUSR1, &sa, NULL );
	sigaction( SIGUSR2, &sa, NULL );
	atexit(clean_exit);

	/* Raise the rlimits in case we're being started from a shell
         * with restrictions. Not a fatal error.  */
	limit.rlim_cur = RLIM_INFINITY;
	limit.rlim_max = RLIM_INFINITY;
	setrlimit(RLIMIT_FSIZE, &limit);
	setrlimit(RLIMIT_CPU, &limit);

	/* Load the Configuration File */
	if (load_config(&config))
		return 6;

	if (config.priority_boost != 0) {
		errno = 0;
		(void) nice((int)-config.priority_boost);
		if (errno) {
			audit_msg(LOG_ERR, "Cannot change priority (%s)", 
					strerror(errno));
			return 1;
		}
	} 
	
	/* Daemonize or stay in foreground for debugging */
	if (config.daemonize == D_BACKGROUND) {
		if (daemon(0,0) != 0) {
			audit_msg(LOG_ERR, "Cannot daemonize (%s)", strerror(errno));
			return 1;
		}
		openlog("auditd", LOG_PID, LOG_DAEMON);
	}

	/* Init netlink */
	if ((fd = audit_open()) < 0) {
        	audit_msg(LOG_ERR, "Cannot open netlink audit socket");
		return 1;
	}

	/* Init the event handler thread */
	write_pid();
	if (init_event(&config))
		return 1;

	/* Write message to log that we are alive */
	{
		char start[64];
		const char *fmt = audit_lookup_format((int)config.log_format);
		if (fmt == NULL)
			fmt = "UNKNOWN";
		snprintf(start, sizeof(start), 
			"auditd start, ver=%s, format=%s, uid=%u", VERSION, 
			fmt, audit_getloginuid());
		if (send_audit_event(AUDIT_DAEMON_START, start)) {
        		audit_msg(LOG_ERR, "Cannot send start message");
			return 1;
		}
	}

	/* Tell the kernel we are alive */
	if (audit_set_pid(fd, getpid(), WAIT_YES) < 0) {
		char emsg[48];
		snprintf(emsg, sizeof(emsg), "auditd error halt, uid=%u",
			audit_getloginuid());
		send_audit_event(AUDIT_DAEMON_ABORT, emsg);
		audit_msg(LOG_ERR, "Unable to set audit pid, exiting");
		return 1;
	}

	if (audit_set_enabled(fd, 1) < 0) {
		char emsg[48];
		snprintf(emsg, sizeof(emsg), "auditd error halt, uid=%u",
			audit_getloginuid());
		send_audit_event(AUDIT_DAEMON_ABORT, emsg);
		audit_msg(LOG_ERR, "Unable to enable auditing, exiting");
		return 1;
	}
	audit_msg(LOG_NOTICE, "Init complete, audit pid set to: %u", getpid());

	for (;;) {
		struct timeval tv;
		int    retval;
       
		if (rep == NULL) { 
			if ((rep = malloc(sizeof(*rep))) == NULL) {
				char emsg[48];
				snprintf(emsg, sizeof(emsg), 
					"auditd error halt, uid=%u",
					audit_getloginuid());
				send_audit_event(AUDIT_DAEMON_ABORT, emsg);
				audit_msg(LOG_ERR, 
					"Cannot allocate audit reply, exiting");
				return 1;
			}
		}
		
		/* This will pause for 30 seconds */
		do {
			tv.tv_sec = 30;
			tv.tv_usec = 0;
			FD_ZERO(&read_mask);
			FD_SET(fd, &read_mask);
			retval = select(fd+1, &read_mask, NULL, NULL, &tv);
		} while (retval == -1 && errno == EINTR && !stop);

		if (!stop && retval > 0) {
			if (audit_get_reply(fd, &rep->reply, 
					GET_REPLY_NONBLOCKING, 0) > 0) {
				switch (rep->reply.type)
				{	/* For now dont process these */
					case NLMSG_NOOP:
					case NLMSG_DONE:
					case NLMSG_ERROR:
					case AUDIT_GET: /* Or these */
					case AUDIT_LIST:
						break;
					default:
						enqueue_event(rep);
						rep = 0;
						break;
				}
			}
		}
		if (stop) {
			/* Write message to log that we are going down */
			int rc;

			rc = audit_request_signal_info(fd);
			if (rc > 0) {
				struct audit_reply trep;

				rc = get_reply(fd, &trep);
				if (rc > 0) {
					char txt[64];
					snprintf(txt, sizeof(txt),
			    	    "auditd normal halt, sending pid=%d uid=%u",
						trep.signal_info->pid, 
						trep.signal_info->uid);
					send_audit_event(AUDIT_DAEMON_END, txt);
				} 
			} 
			if (rc <= 0)
			    send_audit_event(AUDIT_DAEMON_END, 
				"auditd normal halt, sending pid=? uid=?");
			audit_close(fd);
			fd = -1;
			free(rep);
			break;
		}
	}

	/* We are going down. Give the event thread a chance to shutdown.
	   Just in case it hangs, set a timer to get us out of trouble. */
	sa.sa_flags = 0 ;
	sa.sa_handler = thread_killer;
	sigemptyset( &sa.sa_mask ) ;
	sigaction( SIGALRM, &sa, NULL );
	shutdown_events();
	free((void *)config.log_file);

	return 0;
}

/*
 * A clean exit means : 
 * 1) we log that we are going down
 * 2) deregister with kernel
 * 3) close the netlink socket
 */
static void clean_exit(void)
{
	audit_msg(LOG_INFO, "The audit daemon is exiting.");
	if (fd >= 0) {
		audit_set_pid(fd, 0, WAIT_NO);
		audit_close(fd);
	}
	if (pidfile)
		unlink(pidfile);
	closelog();
}

/*
 * This function is used to get the reply for term info.
 * Returns 1 on success & -1 on failure.
 */
static int get_reply(int fd, struct audit_reply *rep)
{
        int rc, i;
        int timeout = 30; /* tenths of seconds */

	for (i = 0; i < timeout; i++) {
		struct timespec t;

		t.tv_sec  = 0;
		t.tv_nsec = 100000000; /* .1 second */
		nanosleep(&t, NULL);
		rc = audit_get_reply(fd, rep, 
			GET_REPLY_NONBLOCKING, 0);
		if (rc > 0) {
			/* If we get done or error, break out */
			if (rep->type == NLMSG_DONE || rep->type == NLMSG_ERROR)
				break;

			/* If its not what we are expecting, keep looping */
			if (rep->type == AUDIT_SIGNAL_INFO)
				return 1;
		}
	}
	return -1;
}

