/*
* ausearch-match.c - Minimal linked list library
* Copyright (c) 2005 Red Hat Inc., Durham, North Carolina.
* All Rights Reserved. 
*
* This software may be freely redistributed and/or modified under the
* terms of the GNU General Public License as published by the Free
* Software Foundation; either version 2, or (at your option) any
* later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING. If not, write to the
* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Authors:
*   Steve Grubb <sgrubb@redhat.com>
*/

#include "config.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/socket.h>
#include <netdb.h>
#include "libaudit.h"
#include "ausearch-llist.h"
#include "ausearch-options.h"


static int strmatch(const char *needle, const char *haystack);
static int user_match(llist *l);
static int group_match(llist *l);
static int extract_search_items(llist *l);
static int parse_syscall(const lnode *n, search_items *s);
static int parse_path(const lnode *n, search_items *s);
static int parse_user(const lnode *n, search_items *s);
static int parse_login(const lnode *n, search_items *s);
static int parse_daemon(const lnode *n, search_items *s);
static int parse_kconfig(const lnode *n, search_items *s);
static int parse_fs_watch(const lnode *n, search_items *s);
static int parse_sockaddr(const lnode *n, search_items *s);

/*
 * This function performs that matching of search params with the record.
 * It returns 1 on a match, and 0 if no match.
 */
int match(llist *l)
{
	// Are we within time range?
	if (start_time == 0 || l->e.sec >= start_time) {
		if (end_time == 0 || l->e.sec <= end_time) {
			if (event_id == -1 || event_id == l->e.serial) {

				// OK - do the heavier checking
				if (extract_search_items(l)) 
					return 0;

				// perform additional tests for the field
				if (user_match(l) == 0)
					return 0;
				if (group_match(l) == 0)
					return 0;
				if ((event_pid != -1) && 
						(event_pid != l->s.pid))
					return 0;
				if ((event_syscall != -1) &&
						(event_syscall != l->s.syscall))
					return 0;

				if ((event_success != -1) &&
						(event_success != l->s.success))
					return 0;
				// event_type requires looking at each item
				if (event_type != -1) {
					int found = 0;
					const lnode *n;
					list_first(l);
					n = list_get_cur(l);
					do {
						if (event_type == n->type) {
							found = 1;
							break;
						}
					} while ((n = list_next(l)));
					if (!found)
						return 0;
				}

				// Done all the easy compares, now do the 
				// string searches.
				if (event_filename) {
					if (l->s.filename == NULL)
						return 0;
					if (strmatch(event_filename,
						l->s.filename) == 0)
						return 0;
				}
				if (event_hostname) {
					if (l->s.hostname == NULL)
						return 0;
					if (strmatch(event_hostname, 
						l->s.hostname) == 0)
						return 0; 
				}				
				if (event_terminal) {
					if (l->s.terminal == NULL)
						return 0;
					if (strmatch(event_terminal, 
						l->s.terminal) == 0)
						return 0; 
				}				
				if (event_exe) {
					if (l->s.exe == NULL)
						return 0;
					if (strmatch(event_exe, 
						l->s.exe) == 0)
						return 0; 
				}				
				return 1;
			}
		}
	}
	return 0;
}

/*
 * This function compares strings. It returns a 0 if no match and a 1 if
 * there is a match 
 */
static int strmatch(const char *needle, const char *haystack)
{
	if (event_exact_match) {
		if (strcmp(haystack, needle) != 0)
			return 0; 
	} else {
		if (strstr(haystack, needle) == NULL)
			return 0; 
	}
	return 1;
}

/*
 * This function compares user id's. It returns a 0 if no match and a 1 if
 * there is a match 
 */
static int user_match(llist *l)
{
	if (event_ua) {
		// This will "or" the user tests
		if (event_uid == l->s.uid)
			return 1;
		if (event_euid == l->s.euid)
			return 1;
		if (event_loginuid == l->s.loginuid)
			return 1;
		return 0;
	} else {
		// This will "and" the user tests
		if ((event_uid != -1) && (event_uid != l->s.uid))
			return 0;
		if ((event_euid != -1) &&(event_euid != l->s.euid))
			return 0;
		if ((event_loginuid != -1) &&
				(event_loginuid != l->s.loginuid))
			return 0;
	}
	return 1;
}

/*
 * This function compares group id's. It returns a 0 if no match and a 1 if
 * there is a match 
 */
static int group_match(llist *l)
{
	if (event_ga) {
		// This will "or" the group tests
		if (event_gid == l->s.gid)
			return 1;
		if (event_egid == l->s.egid)
			return 1;
		return 0;
	} else {
		// This will "and" the group tests
		if ((event_gid != -1) && (event_gid != l->s.gid))
			return 0;
		if ((event_egid != -1) &&(event_egid != l->s.egid))
			return 0;
	}
	return 1;
}

/*
 * This function will take a pointer to a 2 byte Ascii character buffer and 
 * return the actual hex value.
 */
static unsigned char x2c(unsigned char *buf)
{
	static const unsigned char AsciiArray[17] = "0123456789ABCDEF";
        unsigned char *ptr;
        unsigned char total=0;

        ptr = strchr(AsciiArray, (char)toupper(buf[0]));
        if (ptr)
                total = (char)(((ptr-AsciiArray) & 0x0F)<<4);
        ptr = strchr(AsciiArray, (char)toupper(buf[1]));
        if (ptr)
                total += (char)((ptr-AsciiArray) & 0x0F);

        return total;
}

static char *unescape(char *buf)
{
	int len, i;
	char saved, *ptr = buf, *str;

	/* Find the end of the name */
	while (isxdigit(*ptr))
		ptr++;
	saved = *ptr;
	*ptr = 0;
	str = strdup(buf);
	*ptr = saved;

	/* We can get away with this since the buffer is 2 times
	 * bigger than what we are putting there.
	 */
	len = strlen(str);
	ptr = str;
	for (i=0; i<len; i+=2) {
		*ptr = x2c(&str[i]);
		ptr++;
	}
	*ptr = 0;
	return str;
}

/*
 * This function will take the list and extract the searchable fields from it.
 * It returns 0 on success and 1 on failure.
 */
static int extract_search_items(llist *l)
{
	int ret = 0;
	lnode *n;
	search_items *s = &l->s;
	list_first(l);
	n = list_get_cur(l);
	if (n) {
		do {
			char *ptr, *str, *term;
			// get type
			str = strstr(n->message, "type=");
			if (str == NULL)
				return 1;
			ptr = str+5;
			term = strchr(ptr, ' ');
			if (term == NULL)
				return 2;
			*term = 0;
			n->type = audit_name_to_msg_type(ptr);
			if (n->type < 0) {
				fprintf(stderr, "Error converting type:%s\n", 
					ptr);
				return 3;
			}
			*term = ' ';

			switch (n->type) {
			case AUDIT_SYSCALL:
				ret = parse_syscall(n, s);
				break;
			case AUDIT_PATH:
				ret = parse_path(n, s);
				break;
			case AUDIT_FS_WATCH:
				ret = parse_fs_watch(n, s);
				break;
			case AUDIT_USER:
			case AUDIT_FIRST_USER_MSG...AUDIT_LAST_USER_MSG:
				ret = parse_user(n, s);
				break;
			case AUDIT_SOCKADDR:
				ret = parse_sockaddr(n, s);
				break;
			case AUDIT_LOGIN:
				ret = parse_login(n, s);
				break;
			case AUDIT_DAEMON_START:
			case AUDIT_DAEMON_END:
			case AUDIT_DAEMON_ABORT:
			case AUDIT_DAEMON_CONFIG:
				ret = parse_daemon(n, s);
				break;
			case AUDIT_CONFIG_CHANGE:
				ret = parse_kconfig(n, s);
				break;
			case AUDIT_KERNEL:
			case AUDIT_IPC:
			case AUDIT_AVC:
			case AUDIT_SELINUX_ERR:
				// Nothing to parse
				break;
			default:
				break;
			}
		} while ((n=list_next(l)) && ret==0);
	}
	return ret;
}

static int parse_syscall(const lnode *n, search_items *s)
{
	char *ptr, *str, *term;

	// get syscall
	str = strstr(n->message, "syscall=");
	if (str == NULL)
		return 1;
	ptr = str + 8;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 2;
	*term = 0;
	errno = 0;
	s->syscall = (int)strtoul(ptr, NULL, 10);
	if (errno)
		return 3;
	// get success
	*term = ' ';
	str = strstr(term, "success=");
	if (str == NULL)
		return 4;
	ptr = str + 8;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 5;
	*term = 0;
	if (strcmp(ptr, "yes") == 0)
		s->success = 1;
	else
		s->success = 0;
	// pid
	*term = ' ';
	str = strstr(term, "pid=");
	if (str == NULL)
		return 6;
	ptr = str + 4;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 7;
	*term = 0;
	errno = 0;
	s->pid = strtoul(ptr, NULL, 10);
	if (errno)
		return 8;
	// loginuid
	*term = ' ';
	str = strstr(term, "auid=");
	if (str == NULL) {
		str = strstr(term, "loginuid=");
		if (str == NULL)
			return 9;
		ptr = str + 9;
	} else
		ptr = str + 5;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 10;
	*term = 0;
	s->loginuid = strtoul(ptr, NULL, 10);
	if (errno)
		return 11;
	// uid
	*term = ' ';
	str = strstr(term, "uid=");
	if (str == NULL)
		return 12;
	ptr = str + 4;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 13;
	*term = 0;
	s->uid = strtoul(ptr, NULL, 10);
	if (errno)
		return 14;
	// gid
	*term = ' ';
	str = strstr(term, "gid=");
	if (str == NULL)
		return 15;
	ptr = str + 4;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 16;
	*term = 0;
	s->gid = strtoul(ptr, NULL, 10);
	if (errno)
		return 17;
	// euid
	*term = ' ';
	str = strstr(term, "euid=");
	if (str == NULL)
		return 18;
	ptr = str + 5;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 19;
	*term = 0;
	s->euid = strtoul(ptr, NULL, 10);
	if (errno)
		return 20;
	// egid
	*term = ' ';
	str = strstr(term, "egid=");
	if (str == NULL)
		return 21;
	ptr = str + 5;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 22;
	*term = 0;
	s->egid = strtoul(ptr, NULL, 10);
	if (errno)
		return 23;
	*term = ' ';
	if (event_exe) {
		// dont do this search unless needed
		str = strstr(n->message, "exe=");
		if (str) {
			str += 4;
			if (*str == '"') {
				str++;
				term = strchr(str, '"');
				if (term == NULL)
					return 24;
				*term = 0;
				s->exe = strdup(str);
				*term = '"';
			} else 
				s->exe = unescape(str);
		}
	}
	return 0;
}

static int parse_path(const lnode *n, search_items *s)
{
	char *str, *term;

	if (event_filename) {
	// dont do this search unless needed
		str = strstr(n->message, " name=");
		if (str) {
			str += 6;
			if (*str == '"') {
				/* string is normal */
				str++;
				term = strchr(str, '"');
				if (term == NULL)
					return 1;
				*term = 0;
				s->filename = strdup(str);
				*term = '"';
			} else 
				s->filename = unescape(str);
		}
	}
	return 0;
}

static int parse_user(const lnode *n, search_items *s)
{
	char *ptr, *str, *term, saved;

	// get pid
	str = strstr(n->message, "pid=");
	if (str == NULL)
		return 1;
	ptr = str + 4;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 2;
	*term = 0;
	errno = 0;
	s->pid = strtoul(ptr, NULL, 10);
	if (errno)
		return 3;
	*term = ' ';
	// get uid
	str = strstr(term, "uid=");
	if (str == NULL)
		return 4;
	ptr = str + 4;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 5;
	*term = 0;
	s->uid = strtoul(ptr, NULL, 10);
	if (errno)
		return 6;
	*term = ' ';
	// get loginuid
	*term = ' ';
	str = strstr(term, "auid=");
	if (str == NULL) { // Try the older one
		str = strstr(term, "loginuid=");
		if (str == NULL) 
			return 7;
		ptr = str + 9;
	} else
		ptr = str + 5;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 8;
	*term = 0;
	s->loginuid = strtoul(ptr, NULL, 10);
	if (errno)
		return 9;
	*term = ' ';
	// get uid
	str = strstr(term, "uid=");
	if (str != NULL) {
		ptr = str + 4;
		term = strchr(ptr, '\'');
		if (term == NULL) {
			term = strchr(ptr, ' ');
			if (term == NULL)
				return 10;
		}
		saved = *term;
		*term = 0;
		s->uid = strtoul(ptr, NULL, 10);
		if (errno)
			return 11;
		*term = saved;
	}
	// get hostname
	if (event_hostname) {
	// dont do this search unless needed
		str = strstr(n->message, "hostname=");
		if (str) {
			str += 9;
			term = strchr(str, ',');
			if (term == NULL)
				return 12;
			*term = 0;
			s->hostname = strdup(str);
			*term = ',';
		}
	}
	if (event_terminal) {
		// dont do this search unless needed
		str = strstr(n->message, "terminal=");
		if (str) {
			str += 9;
			term = strchr(str, ' ');
			if (term == NULL)
				return 13;
			*term = 0;
			s->terminal = strdup(str);
			*term = ' ';
		}
	}
	if (event_exe) {
	// dont do this search unless needed
		str = strstr(n->message, "exe=");
		if (str) {
			str += 4;
			if (*str == '"') {
				str++;
				term = strchr(str, '"');
				if (term == NULL)
					return 14;
				*term = 0;
				s->exe = strdup(str);
				*term = '"';
			} else 
				s->exe = unescape(str);
		}
	}
	return 0;
}

static int parse_login(const lnode *n, search_items *s)
{
	char *ptr, *str, *term;

	// get pid
	str = strstr(n->message, "pid=");
	if (str == NULL)
		return 1;
	ptr = str + 4;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 2;
	*term = 0;
	errno = 0;
	s->pid = strtoul(ptr, NULL, 10);
	if (errno)
		return 3;
	// get uid
	*term = ' ';
	str = strstr(term, "uid=");
	if (str == NULL)
		return 4;
	ptr = str + 4;
	term = strchr(ptr, ' ');
	if (term == NULL)
		return 5;
	*term = 0;
	s->uid = strtoul(ptr, NULL, 10);
	if (errno)
		return 6;
	// get loginuid
	*term = ' ';
	str = strstr(term, "new auid=");
	if (str == NULL) {
		str = strstr(term, "new loginuid=");
		if (str == NULL)
			return 7;
		ptr = str + 13;
	} else
		ptr = str + 9;
	term = strchr(ptr, ' ');
	if (term)
		*term = 0;
	s->loginuid = strtoul(ptr, NULL, 10);
	if (errno)
		return 8;
	if (term)
		*term = ' ';
	return 0;
}

static int parse_daemon(const lnode *n, search_items *s)
{
	char *ptr, *str, *term;

	// uid
	str = strstr(n->message, "uid=");
	if (str == NULL)
		return 1;
	ptr = str + 4;
	term = strchr(ptr, ',');
	if (term == NULL)
		return 2;
	*term = 0;
	errno = 0;
	s->uid = strtoul(ptr, NULL, 10);
	if (errno)
		return 3;
	// pid
	*term = ' ';
	str = strstr(term, "pid=");
	if (str == NULL)
		return 4;
	ptr = str + 4;
	term = strchr(ptr, ' ');
	if (term)
		*term = 0;
	s->pid = strtoul(ptr, NULL, 10);
	if (errno)
		return 5;
	if (term)
		*term = ' ';
	return 0;
}

static int parse_kconfig(const lnode *n, search_items *s)
{
	char *ptr, *str, *term;

	// get loginuid
	str = strstr(n->message, "auid ");
	if (str == NULL)
		return 1;
	ptr = str + 5;
	term = strchr(ptr, ' ');
	if (term)
		*term = 0;
	errno = 0;
	s->loginuid = strtoul(ptr, NULL, 10);
	if (errno)
		return 2;
	if (term)
		*term = ' ';
	return 0;
}

static int parse_fs_watch(const lnode *n, search_items *s)
{
	char *str, *term;

	if (event_filename) {
	// dont do this search unless needed
		str = strstr(n->message, "watch=");
		if (str) {
			str += 6;
			if (*str == '"') {
				str++;
				term = strchr(str, '"');
				if (term == NULL)
					return 1;
				*term = 0;
				s->filename = strdup(str);
				*term = '"';
			} else 
				s->filename = unescape(str);
		}
	}
	return 0;
}

static int parse_sockaddr(const lnode *n, search_items *s)
{
	char *str;

	if (event_hostname) {
		str = strstr(n->message, "saddr=");
		if (str) {
			int len;
			struct sockaddr *saddr;
			char name[NI_MAXHOST];

			str += 6;
			len = strlen(str)/2;
			s->hostname = unescape(str);
			saddr = (struct sockaddr *)s->hostname;
			if (saddr->sa_family == AF_INET) {
				if (len < sizeof(struct sockaddr_in)) {
					fprintf(stderr,
						"sockaddr len too short\n");
					return 1;
				}
				len = sizeof(struct sockaddr_in);
			} else if (saddr->sa_family == AF_INET6) {
				if (len < sizeof(struct sockaddr_in6)) {
					fprintf(stderr,
						"sockaddr6 len too short\n");
					return 2;
				}
				 len = sizeof(struct sockaddr_in6);
			} else {
				// addr family we don't care about
				// should we convert unix domain sockets?
				free(s->hostname);
				s->hostname = NULL;
				return 0;
			}
			if (getnameinfo(saddr, len, name, NI_MAXHOST, 
					NULL, 0, NI_NUMERICHOST) ) {
				free(s->hostname);
				s->hostname = NULL;
			} else {
				free(s->hostname);
				s->hostname = strdup(name);
			}
		}
	}
	return 0;
}

