/* ausearch-options.c -- 
 * Copyright 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *     Debora Velarde <dvelarde@us.ibm.com>
 *     Steve Grubb <sgrubb@redhat.com>
 */

#include "config.h"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#define _XOPEN_SOURCE	/* Needed for strptime */
#define __USE_XOPEN
#include <time.h>
#undef __USE_XOPEN
#undef _XOPEN_SOURCE
#include "ausearch-options.h"
#include "libaudit.h"


/* Global vars that will be accessed by the match model */
time_t start_time = 0, end_time = 0;
unsigned int event_id = -1;
gid_t event_gid = -1, event_egid = -1;
int event_type = -1;
pid_t event_pid = -1;
int event_success = -1, event_exact_match = 0;
uid_t event_uid = -1, event_euid = -1, event_loginuid = -1;
int event_syscall = -1;
int event_ua = 0, event_ga = 0;
const char *event_filename = NULL, *event_hostname = NULL;
const char *event_terminal = NULL;
const char *event_exe = NULL;

struct nv_pair {
    int        value;
    const char *name;
};

enum { S_EVENT, S_FILENAME, S_ALL_GID, S_EFF_GID, S_GID, S_HELP, S_HOSTNAME,
S_MESSAGE_TYPE, S_PID, S_SYSCALL, S_SUCCESS, S_TIME_END, S_TIME_START, 
S_TERMINAL, S_ALL_UID, S_EFF_UID, S_UID, S_LOGINID, S_VERSION, S_EXACT_MATCH,
S_EXECUTABLE };

static struct nv_pair optiontab[] = {
	{ S_EVENT, "-a" },
	{ S_FILENAME, "-f" },
	{ S_ALL_GID, "-ga" },
	{ S_EFF_GID, "-ge" },
	{ S_GID, "-gi" },
	{ S_HELP, "-h" },
	{ S_HOSTNAME, "-hn" },
	{ S_MESSAGE_TYPE, "-m" },
	{ S_PID, "-p" },
	{ S_SYSCALL, "-sc" },
	{ S_SUCCESS, "-sv" },
	{ S_TIME_END, "-te"},
	{ S_TIME_START, "-ts" },
	{ S_TERMINAL, "-tm" },
	{ S_ALL_UID, "-ua" },
	{ S_EFF_UID, "-ue" },
	{ S_UID, "-ui" },
	{ S_LOGINID, "-ul" },
	{ S_VERSION, "-v" },
	{ S_EXACT_MATCH, "-w" },
	{ S_EXECUTABLE, "-x" }
};
#define OPTION_NAMES (sizeof(optiontab)/sizeof(optiontab[0]))


static int audit_lookup_option(const char *name)
{
        int i;

        for (i = 0; i < OPTION_NAMES; i++)
                if (!strcmp(optiontab[i].name, name))
			return optiontab[i].value;
        return -1;
}

static void usage(void)
{
	printf("usage: ausearch [options]\n"
	"       -a  <audit event id>\n"
	"       -f  <file name>\n"
	"       -ga <all group id>\n"
	"       -ge <effective group id>\n"
	"       -gi <group id>\n"
	"       -h\n"
	"       -hn <host name>\n"
	"       -m  <message type>\n"
	"       -p  <process id>\n"
	"       -sc <syscall name>\n"
	"       -sv <success value>\n"
	"       -te [end date] [end time]\n"
	"       -ts [start date] [start time]\n"
	"       -tm <terminal>\n"
	"       -ua <all user id>\n"
	"       -ue <effective user id>\n"
	"       -ui <user id>\n"
	"       -ul <login id>\n"
	"       -v\n"
	"       -w\n"
	"       -x <executable name>\n"
	);
}

void clear_tm(struct tm *t)
{
	t->tm_sec = 0;         /* seconds */
	t->tm_min = 0;         /* minutes */
	t->tm_hour = 0;        /* hours */
	t->tm_mday = 0;        /* day of the month */
	t->tm_mon = 0;         /* month */
	t->tm_year = 0;        /* year */
	t->tm_isdst = 0;       /* DST flag */
}

/* Combine date & time into 1 struct. Results in date. */
void add_tm(struct tm *d, struct tm *t)
{
	time_t dst;
	struct tm *lt;

	d->tm_sec = t->tm_sec;         /* seconds */
	d->tm_min = t->tm_min;         /* minutes */
	d->tm_hour = t->tm_hour;        /* hours */

	/* Now we need to figure out if DST is in effect */
	dst = time(NULL);
	lt = localtime(&dst);
	d->tm_isdst = lt->tm_isdst;
}

int ausearch_time_start(const char *da, const char *ti)
{
/* If da == NULL, use current date */
/* If ti == NULL, then use midnight 00:00:00 */
	int rc = 0;
	struct tm d, t;
	char *ret;

	if (da != NULL) {
		ret = strptime(da, "%x", &d);
		if (ret == NULL) {
			fprintf(stderr, "Invalid start date (%s). Month, Day, and Year are required.\n", da);
			return 1;
		}
		if (*ret != 0) {
			fprintf(stderr, "Error parsing start date (%s)\n", da);
			return 1;
		}
	} else {
		time_t t = time(NULL);
		struct tm *tv = localtime(&t);
		clear_tm(&d);
		d.tm_year = tv->tm_year;
		d.tm_mon = tv->tm_mon;
		d.tm_mday = tv->tm_mday;
		d.tm_isdst = tv->tm_isdst;
	}
	if (ti != NULL) {
		ret = strptime(ti, "%X", &t);
		if (ret == NULL) {
			fprintf(stderr, "Invalid start time (%s). Hour, Minute, and Second are required.\n", ti);
			return 1;
		}
		if (*ret != 0) {
			fprintf(stderr, "Error parsing start time (%s)\n", ti);
			return 1;
		}
	}
	else
		clear_tm(&t);
	add_tm(&d, &t);
	start_time = mktime(&d); 
	if (start_time == -1) {
		fprintf(stderr, "Error converting start time\n");
		rc = -1;
	}
	return rc;
}

int ausearch_time_end(const char *da, const char *ti)
{
/* If date == NULL, use current date */
/* If ti == NULL, use current time */
	int rc = 0;
	struct tm d, t;
	char *ret;

	if (da != NULL) {
		ret = strptime(da, "%x", &d);
		if (ret == NULL) {
			fprintf(stderr, "Invalid end date (%s)\n", da);
			return 1;
		}
		if (*ret != 0) {
			fprintf(stderr, "Error parsing end date (%s)\n", da);
			return 1;
		}
	} else {
		time_t t = time(NULL);
		struct tm *tv = localtime(&t);
		clear_tm(&d);
		d.tm_year = tv->tm_year;
		d.tm_mon = tv->tm_mon;
		d.tm_mday = tv->tm_mday;
		d.tm_isdst = tv->tm_isdst;
	}
	if (ti != NULL) {
		ret = strptime(ti, "%X", &t);
		if (ret == NULL) {
			fprintf(stderr, "Invalid end time (%s)\n", ti);
			return 1;
		}
		if (*ret != 0) {
			fprintf(stderr, "Error parsing end time (%s)\n", ti);
			return 1;
		}
	} else {
		time_t tt = time(NULL);
		struct tm *tv = localtime(&tt);
		clear_tm(&t); 
		t.tm_hour = tv->tm_hour;
		t.tm_min = tv->tm_min;
		t.tm_sec = tv->tm_sec;
		t.tm_isdst = tv->tm_isdst;
	}
	add_tm(&d, &t); 
	end_time = mktime(&d);
	if (end_time == -1) {
		fprintf(stderr, "Error converting end time\n");
		rc = -1;
	}
	return rc;
}

/*
 * This function examines the commandline parameters and sets various
 * search options. It returns a 0 on success and < 0 on failure
 */
int check_params(int count, char *vars[])
{
	int c = 1;
	int retval = 0;
	const char *optarg;

	if (count < 2) {
		usage();
		return -1;
	}
	while (c < count && retval == 0) {
		// Go ahead and point to the next argument
		if (c+1 < count) {
			if (vars[c+1][0] != '-')
				optarg = vars[c+1];
			else
				optarg = NULL;
		} else
			optarg = NULL;

		switch (audit_lookup_option(vars[c])) {
		case S_EVENT:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
			if (isdigit(optarg[0])) {
				errno = 0;
				event_id = strtoul(optarg, NULL, 10);
				if (errno) {
					fprintf(stderr,
					"Illegal value for audit event ID");
					retval = -1;
				}
				c++;
			} else {
				fprintf(stderr, 
			"Audit event id must be a numeric value, was %s\n",
					optarg);
				retval = -1;
			}
			break;
		case S_FILENAME:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
			} else {
				event_filename = strdup(optarg);
        	                if (event_filename == NULL)
                	                retval = -1;
				c++;
			}
			break;
		case S_ALL_GID:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
                	if (isdigit(optarg[0])) {
				errno = 0;
                        	event_gid = strtoul(optarg,NULL,10);
				if (errno) {
                        		fprintf(stderr, 
			"Numeric group ID conversion error (%s) for %s\n",
						strerror(errno), optarg);
                                	retval = -1;
				}
                	} else {
				struct group *gr ;

				gr = getgrnam(optarg) ;
				if (gr == NULL) {
                        		fprintf(stderr, 
				"Group ID is non-numeric and unknown (%s)\n",
						optarg);
					retval = -1;
					break;
				}
				event_gid = gr->gr_gid;
                	}
			event_egid = event_gid;
			event_ga = 1;
			c++;
			break;
		case S_EFF_GID:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
                	if (isdigit(optarg[0])) {
				errno = 0;
                        	event_egid = strtoul(optarg,NULL,10);
				if (errno) {
                        		fprintf(stderr, 
			"Numeric group ID conversion error (%s) for %s\n",
						strerror(errno), optarg);
                                	retval = -1;
				}
                	} else {
				struct group *gr ;

				gr = getgrnam(optarg) ;
				if (gr == NULL) {
                        		fprintf(stderr, 
			"Effective group ID is non-numeric and unknown (%s)\n",
						optarg);
					retval = -1;
					break;
				}
				event_egid = gr->gr_gid;
                	}
			c++;
			break;
		case S_GID:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
                	if (isdigit(optarg[0])) {
				errno = 0;
                        	event_gid = strtoul(optarg,NULL,10);
				if (errno) {
                        		fprintf(stderr, 
			"Numeric group ID conversion error (%s) for %s\n",
						strerror(errno), optarg);
                                	retval = -1;
				}
                	} else {
				struct group *gr ;

				gr = getgrnam(optarg) ;
				if (gr == NULL) {
                        		fprintf(stderr, 
				"Group ID is non-numeric and unknown (%s)\n",
						optarg);
					retval = -1;
					break;
				}
				event_gid = gr->gr_gid;
                	}
			c++;
			break;
		case S_HELP:
			usage();
			exit(0);
			break;
		case S_HOSTNAME:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
			} else {
				event_hostname = strdup(optarg);
				if (event_hostname == NULL)
					retval = -1;
				c++;
			}
			break;
		case S_MESSAGE_TYPE:
	                if (!optarg) {
				fprintf(stderr, "Argument is required\n");
        	                retval = -1;
	                } else {
				if (strcasecmp(optarg, "ALL")) {
					event_type =
						audit_name_to_msg_type(optarg);
					if (event_type < 0) 
                        		        retval = -1;
				}
				c++;
			}
			if (retval < 0) {
				int i;
                        	fprintf(stderr, 
					"Valid message types are: ALL ");
				for (i=AUDIT_USER; i<=AUDIT_KERNEL; i++) {
					const char *name;
					if (i == AUDIT_WATCH_INS) // Skip a few
						i = AUDIT_USER_AUTH;
					name = audit_msg_type_to_name(i);
					if (name)
		                        	fprintf(stderr, "%s ", name);
				}
                        	fprintf(stderr, "\n");
			}
			break;
		case S_PID:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
			if (isdigit(optarg[0])) {
				errno = 0;
				event_pid = strtol(optarg,NULL,10);
				if (errno)
					retval = -1;
				c++;
			} else {
				fprintf(stderr, 
				"Process id must be a numeric value, was %s\n",
					optarg);
				retval = -1;
			}
			break;
		case S_SYSCALL:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
			if (optarg && isdigit(optarg[0])) {
				errno = 0;
				event_syscall = (int)strtoul(optarg, NULL, 10);
				if (errno) {
                        		fprintf(stderr, 
			"Syscall numeric conversion error (%s) for %s\n",
						strerror(errno), optarg);
                                	retval = -1;
				}
			} else {
                                int machine;
                                machine = audit_detect_machine();
                                if (machine < 0) {
                                        fprintf(stderr,
                                            "Error detecting machine type");
                                        retval = -1;
					break;
                                }
				event_syscall = audit_name_to_syscall(optarg, 
					machine);
				if (event_syscall == -1) {
					fprintf(stderr, 
						"Syscall %s not found\n",
						optarg);
                                        retval = -1;
				}
                        }
			c++;
			break;
		case S_SUCCESS:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
                	if ( (strstr(optarg, "yes")!=NULL) || 
					(strstr(optarg, "no")!=NULL) ) {
                        	if (strcmp(optarg, "yes") == 0)
					event_success = 1;
				else
					event_success = 0;
                	} else {
                        	fprintf(stderr, 
					"Success must be 'yes' or 'no'.\n");
                        	retval = -1;
                	}
			c++;
			break;
		case S_TIME_END:
			if (optarg) {
				if ( (c+2 < count) && vars[c+2] && 
					(vars[c+2][0] != '-') ) {
				/* Have both date and time - check order*/
					if (strchr(optarg, ':')) {
						if (ausearch_time_end(vars[c+2],
								 optarg) != 0) 
							retval = -1;
					} else {
						if (ausearch_time_end(optarg, 
								vars[c+2]) != 0)
							retval = -1;
					}
					c++;			
				} else {
					if ( (strchr(optarg, ':')) == NULL) {
						/* Only have date */
						if (ausearch_time_end(optarg,
								NULL) != 0)
							retval = -1;
					} else {
						/* Only have time */
						if (ausearch_time_end(NULL,
								optarg) != 0)
							retval = -1;
					}
				}
				c++;			
				break;
			} else {
				if (ausearch_time_end(NULL, NULL) != 0)
					retval = -1;
				break;
			}
			break;
		case S_TIME_START:
			if (optarg) {
				if ( (c+2 < count) && vars[c+2] && 
					(vars[c+2][0] != '-') ) {
				/* Have both date and time - check order */
					if (strchr(optarg, ':')) {
						if (ausearch_time_start(
							vars[c+2], optarg) != 0)
							retval = -1;
					} else {
						if (ausearch_time_start(optarg, 
								vars[c+2]) != 0)
							retval = -1;
					}
					c++;
				} else {
					if ( strchr(optarg, ':') == NULL) {
						/* Only have date */
						if (ausearch_time_start(optarg,
							"00:00:01") != 0)
							retval = -1;
					} else {
						/* Only have time */
						if (ausearch_time_start(NULL,
								optarg) != 0)
							retval = -1;
					}
				}
				c++;
				break;
			}
			fprintf(stderr, 
				"-ts requires either date and/or time\n");
			retval = -1;
			break;
		case S_TERMINAL:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
                	} else { 
				event_terminal = strdup(optarg);
				if (event_terminal == NULL)
        	                        retval = -1;
				c++;
			}
			break;
		case S_UID:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
                	if (isdigit(optarg[0])) {
				errno = 0;
                        	event_uid = strtoul(optarg,NULL,10);
				if (errno) {
                        		fprintf(stderr, 
			"Numeric user ID conversion error (%s) for %s\n",
						strerror(errno), optarg);
                                	retval = -1;
				}
                	} else {
				struct passwd *pw ;

				pw = getpwnam(optarg) ;
				if (pw == NULL) {
                        		fprintf(stderr, 
			"Effective user ID is non-numeric and unknown (%s)\n",
						optarg);
					retval = -1;
					break;
				}
				event_uid = pw->pw_uid;
                	}
			c++;
			break;
		case S_EFF_UID:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
                	if (isdigit(optarg[0])) {
				errno = 0;
                        	event_euid = strtoul(optarg,NULL,10);
				if (errno) {
                        		fprintf(stderr, 
			"Numeric user ID conversion error (%s) for %s\n",
						strerror(errno), optarg);
                                	retval = -1;
				}
                	} else {
				struct passwd *pw ;

				pw = getpwnam(optarg) ;
				if (pw == NULL) {
                        		fprintf(stderr, 
				"User ID is non-numeric and unknown (%s)\n",
						optarg);
					retval = -1;
					break;
				}
				event_euid = pw->pw_uid;
                	}
			c++;
			break;
		case S_ALL_UID:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
                	if (isdigit(optarg[0])) {
				errno = 0;
                        	event_uid = strtoul(optarg,NULL,10);
				if (errno) {
                        		fprintf(stderr, 
			"Numeric user ID conversion error (%s) for %s\n",
						strerror(errno), optarg);
                                	retval = -1;
				}
                	} else {
				struct passwd *pw ;

				pw = getpwnam(optarg) ;
				if (pw == NULL) {
                        		fprintf(stderr, 
				"User ID is non-numeric and unknown (%s)\n",
						optarg);
					retval = -1;
					break;
				}
				event_uid = pw->pw_uid;
                	}
			event_ua = 1;
			event_euid = event_uid;
			event_loginuid = event_uid;
			c++;
			break;
		case S_LOGINID:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
				break;
			}
                        if (isdigit(optarg[0])) {
				errno = 0;
                        	event_loginuid = strtoul(optarg,NULL,10);
				if (errno) {
                        		fprintf(stderr, 
			"Numeric user ID conversion error (%s) for %s\n",
						strerror(errno), optarg);
                                        retval = -1;
				}
                        } else {
				struct passwd *pw ;

				pw = getpwnam(optarg) ;
				if (pw == NULL) {
                        		fprintf(stderr, 
			"Login user ID is non-numeric and unknown (%s)\n",
						optarg);
					retval = -1;
					break;
				}
				event_loginuid = pw->pw_uid;
                        }
			c++;
			break;
		case S_VERSION:
	                printf("ausearch version %s\n", VERSION);
			exit(0);
			break;
		case S_EXACT_MATCH:
			event_exact_match=1;
			break;
		case S_EXECUTABLE:
			if (!optarg) {
				fprintf(stderr, "Argument is required\n");
				retval = -1;
	               	} else { 
				event_exe = strdup(optarg);
				if (event_exe == NULL)
       	                	        retval = -1;
				c++;
			}
			break;
		default:
			fprintf(stderr, "%s is an unsupported option\n", 
				vars[c]);
			retval = -1;
			break;
		}
		c++;
	}

	return retval;
}

