/* lookup_table.c -- 
 * Copyright 2004, 2005 Red Hat Inc., Durham, North Carolina.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors:
 *      Steve Grubb <sgrubb@redhat.com>
 *      Rickard E. (Rik) Faith <faith@redhat.com>
 */

#include "config.h"
#include <string.h>
#include <unistd.h>

#include "libaudit.h"
#define __KERNEL__
#include <sys/syscall.h>
#undef __KERNEL__

struct transtab {
    int        value;
    const char *name;
};

struct int_transtab {
    int        key;
    unsigned int  lvalue;
};        

static struct transtab i386_syscalltab[] = {
    #include "i386_table.h"
};
#define AUDIT_I386_SYSCALL_NAMES (sizeof(i386_syscalltab)/sizeof(i386_syscalltab[0]))

static struct transtab x86_64_syscalltab[] = {
    #include "x86_64_table.h"
};
#define AUDIT_X86_64_SYSCALL_NAMES (sizeof(x86_64_syscalltab)/sizeof(x86_64_syscalltab[0]))

static struct transtab ppc64_syscalltab[] = {
    #include "ppc64_table.h"
};
#define AUDIT_PPC64_SYSCALL_NAMES (sizeof(ppc64_syscalltab)/sizeof(ppc64_syscalltab[0]))


static struct transtab ppc_syscalltab[] = {
    #include "ppc_table.h"
};
#define AUDIT_PPC_SYSCALL_NAMES (sizeof(ppc_syscalltab)/sizeof(ppc_syscalltab[0]))

static struct transtab s390x_syscalltab[] = {
    #include "s390x_table.h"
};
#define AUDIT_S390X_SYSCALL_NAMES (sizeof(s390x_syscalltab)/sizeof(s390x_syscalltab[0]))

static struct transtab s390_syscalltab[] = {
    #include "s390_table.h"
};
#define AUDIT_S390_SYSCALL_NAMES (sizeof(s390_syscalltab)/sizeof(s390_syscalltab[0]))

static struct transtab ia64_syscalltab[] = {
    #include "ia64_table.h"
};
#define AUDIT_IA64_SYSCALL_NAMES (sizeof(ia64_syscalltab)/sizeof(ia64_syscalltab[0]))

static struct transtab fieldtab[] = {
    { AUDIT_PID,      "pid"      },
    { AUDIT_UID,      "uid"      },
    { AUDIT_EUID,     "euid"     },
    { AUDIT_SUID,     "suid"     },
    { AUDIT_FSUID,    "fsuid"    },
    { AUDIT_GID,      "gid"      },
    { AUDIT_EGID,     "egid"     },
    { AUDIT_SGID,     "sgid"     },
    { AUDIT_FSGID,    "fsgid"    },
    { AUDIT_LOGINUID, "loginuid" },
    { AUDIT_PERS,     "pers"     },
    { AUDIT_ARCH,     "arch"     },

    { AUDIT_DEVMAJOR, "devmajor" },
    { AUDIT_DEVMINOR, "devminor" },
    { AUDIT_INODE,    "inode"    },
    { AUDIT_EXIT,     "exit"     },
    { AUDIT_SUCCESS,  "success"  },
    { AUDIT_ARG0,     "a0"       },
    { AUDIT_ARG1,     "a1"       },
    { AUDIT_ARG2,     "a2"       },
    { AUDIT_ARG3,     "a3"       },
};
#define AUDIT_FIELD_NAMES (sizeof(fieldtab)/sizeof(fieldtab[0]))

static struct transtab flagtab[] = {
    { AUDIT_PER_TASK, "task"     },
    { AUDIT_AT_ENTRY, "entry"    },
    { AUDIT_AT_EXIT,  "exit"     }
};
#define AUDIT_FLAG_NAMES (sizeof(flagtab)/sizeof(flagtab[0]))

static struct transtab actiontab[] = {
    { AUDIT_NEVER,    "never"    },
    { AUDIT_POSSIBLE, "possible" },
    { AUDIT_ALWAYS,   "always"   }
};
#define AUDIT_ACTION_NAMES (sizeof(actiontab)/sizeof(actiontab[0]))

/* 
 * This table is arranged based on statistics so more frequent items
 * are at the top.
 */
static struct transtab msg_typetab[] = {
    { AUDIT_KERNEL,		"KERNEL"      },
    { AUDIT_USER,		"USER"	      },
    { AUDIT_LOGIN,		"LOGIN"       },
    { AUDIT_SIGNAL_INFO,	"SIGNAL_INFO" },
    { AUDIT_SET,		"SET"	      },
    { AUDIT_ADD,		"ADD"	      },
    { AUDIT_LIST,		"LIST"	      },
    { AUDIT_DEL,		"DEL"	      },
    { AUDIT_GET,		"GET"	      },
    { AUDIT_USER_AUTH,		"USER_AUTH"	},
    { AUDIT_USER_ACCT,		"USER_ACCT"	},
    { AUDIT_USER_MGMT,		"USER_MGMT"	},
    { AUDIT_CRED_ACQ,		"CRED_ACQ"	},
    { AUDIT_CRED_DISP,		"CRED_DISP"	},
    { AUDIT_USER_START,		"USER_START"	},
    { AUDIT_USER_END,		"USER_END"	},
    { AUDIT_USER_AVC,		"USER_AVC"	},
    { AUDIT_USER_CHAUTHTOK,	"USER_CHAUTHTOK"},
    { AUDIT_USER_ERR,		"USER_ERR"	},
    { AUDIT_CRED_REFR,		"CRED_REFR"	},
    { AUDIT_DAEMON_START,	"DAEMON_START"	},
    { AUDIT_DAEMON_END,		"DAEMON_END"	},
    { AUDIT_DAEMON_ABORT,	"DAEMON_ABORT"	},
    { AUDIT_DAEMON_CONFIG,	"DAEMON_CONFIG"	},
    { AUDIT_SYSCALL,		"SYSCALL"	},
    { AUDIT_FS_WATCH,		"FS_WATCH"	},
    { AUDIT_PATH,		"PATH"		},
    { AUDIT_IPC,		"IPC"		},
    { AUDIT_SOCKETCALL,		"SOCKETCALL"	},
    { AUDIT_CONFIG_CHANGE,	"CONFIG_CHANGE" },
    { AUDIT_SOCKADDR,		"SOCKADDR"	},
    { AUDIT_AVC,		"AVC"		},
    { AUDIT_SELINUX_ERR,	"SELINUX_ERR"	},
    { AUDIT_AVC_PATH,		"AVC_PATH"	}
};
#define AUDIT_MSG_TYPE_NAMES (sizeof(msg_typetab)/sizeof(msg_typetab[0]))

static struct transtab machinetab[] = {
    { MACH_X86,     "i386"   },
    { MACH_X86,     "i486"   },
    { MACH_X86,     "i586"   },
    { MACH_X86,     "i686"   },
    { MACH_86_64,   "x86_64" },
    { MACH_IA64,    "ia64"   },
    { MACH_PPC64,   "ppc64"  },
    { MACH_PPC,     "ppc"    },
    { MACH_S390X,   "s390x"  },
    { MACH_S390,    "s390"   }
};
#define AUDIT_MACHINE_NAMES (sizeof(machinetab)/sizeof(machinetab[0]))

static struct int_transtab elftab[] = {
    { MACH_X86,     AUDIT_ARCH_I386   },
    { MACH_86_64,   AUDIT_ARCH_X86_64 },
    { MACH_IA64,    AUDIT_ARCH_IA64   },
    { MACH_PPC64,   AUDIT_ARCH_PPC64  },
    { MACH_PPC,     AUDIT_ARCH_PPC    },
    { MACH_S390X,   AUDIT_ARCH_S390X  },
    { MACH_S390,    AUDIT_ARCH_S390   }
};
#define AUDIT_ELF_NAMES (sizeof(elftab)/sizeof(elftab[0]))

static int audit_lookup_name(struct transtab *table, int length,
                             const char *name)
{
	int i;
    
	for (i = 0; i < length; i++)
		if (!strcmp(table[i].name, name)) return table[i].value;
	return -1;
}

static const char *audit_lookup_number(struct transtab *table, int length,
                                       int number)
{
	int i;
    
	for (i = 0; i < length; i++)
		if (table[i].value == number) return table[i].name;
	return NULL;
}

int audit_name_to_field(const char *field)
{
	return audit_lookup_name(fieldtab, AUDIT_FIELD_NAMES, field);
}

const char *audit_field_to_name(int field)
{
	return audit_lookup_number(fieldtab, AUDIT_FIELD_NAMES, field);
}

int audit_name_to_syscall(const char *sc, int machine)
{
	switch (machine)
	{
		case MACH_X86:
    			return audit_lookup_name(i386_syscalltab, 
					AUDIT_I386_SYSCALL_NAMES, sc);
		case MACH_86_64:
    			return audit_lookup_name(x86_64_syscalltab, 
					AUDIT_X86_64_SYSCALL_NAMES, sc);
		case MACH_IA64:
    			return audit_lookup_name(ia64_syscalltab, 
					AUDIT_IA64_SYSCALL_NAMES, sc);
		case MACH_PPC64:
    			return audit_lookup_name(ppc64_syscalltab, 
					AUDIT_PPC64_SYSCALL_NAMES, sc);
		case MACH_PPC:
    			return audit_lookup_name(ppc_syscalltab, 
					AUDIT_PPC_SYSCALL_NAMES, sc);
		case MACH_S390X:
    			return audit_lookup_name(s390x_syscalltab, 
					AUDIT_S390X_SYSCALL_NAMES, sc);
		case MACH_S390:
    			return audit_lookup_name(s390_syscalltab, 
					AUDIT_S390_SYSCALL_NAMES, sc);
	}
	return -1;
}

const char *audit_syscall_to_name(int sc, int machine)
{
	switch (machine)
	{
		case MACH_X86:
    			return audit_lookup_number(i386_syscalltab, 
					AUDIT_I386_SYSCALL_NAMES, sc);
		case MACH_86_64:
    			return audit_lookup_number(x86_64_syscalltab, 
					AUDIT_X86_64_SYSCALL_NAMES, sc);
		case MACH_IA64:
    			return audit_lookup_number(ia64_syscalltab, 
					AUDIT_IA64_SYSCALL_NAMES, sc);
		case MACH_PPC64:
    			return audit_lookup_number(ppc64_syscalltab, 
					AUDIT_PPC64_SYSCALL_NAMES, sc);
		case MACH_PPC:
    			return audit_lookup_number(ppc_syscalltab, 
					AUDIT_PPC_SYSCALL_NAMES, sc);
		case MACH_S390X:
    			return audit_lookup_number(s390x_syscalltab, 
					AUDIT_S390X_SYSCALL_NAMES, sc);
		case MACH_S390:
    			return audit_lookup_number(s390_syscalltab, 
					AUDIT_S390_SYSCALL_NAMES, sc);
	}
	return NULL;
}

int audit_name_to_flag(const char *flag)
{
	return audit_lookup_name(flagtab, AUDIT_FLAG_NAMES, flag);
}

/* FIXME: only returns the first one in the list. */
const char *audit_flag_to_name(int flag)
{
	unsigned int i;
    
	for (i = 0; i < AUDIT_FLAG_NAMES; i++)
		if (flagtab[i].value & flag)
			return flagtab[i].name;
	return NULL;
}

int audit_name_to_action(const char *action)
{
	return audit_lookup_name(actiontab, AUDIT_ACTION_NAMES, action);
}

const char *audit_action_to_name(int action)
{
	return audit_lookup_number(actiontab, AUDIT_ACTION_NAMES, action);
}

int audit_name_to_msg_type(const char *msg_type)
{
	return audit_lookup_name(msg_typetab, AUDIT_MSG_TYPE_NAMES, msg_type);
}

const char *audit_msg_type_to_name(int msg_type)
{
	return audit_lookup_number(msg_typetab, AUDIT_MSG_TYPE_NAMES, msg_type);
}

int audit_name_to_machine(const char *machine)
{
	return audit_lookup_name(machinetab, AUDIT_MACHINE_NAMES, machine);
}

const char *audit_machine_to_name(int machine)
{
	return audit_lookup_number(machinetab, AUDIT_MACHINE_NAMES, machine);
}

unsigned int audit_machine_to_elf(int machine)
{
	unsigned int i;
    
	for (i = 0; i < AUDIT_ELF_NAMES; i++)
		if (elftab[i].key == machine) 
			return elftab[i].lvalue;
	return 0;
}

int audit_elf_to_machine(unsigned int elf)
{
	unsigned int i;
    
	for (i = 0; i < AUDIT_ELF_NAMES; i++)
		if (elftab[i].lvalue == elf) return elftab[i].key;
	return -1;
}

